from .populations import StarPopulation
from .populations import MultipleStarPopulation, ColormatchMultipleStarPopulation
from .populations import Spectroscopic_MultipleStarPopulation
from .populations import BGStarPopulation, BGStarPopulation_TRILEGAL
from .populations import BinaryPopulation, Simulated_BinaryPopulation
from .populations import Raghavan_BinaryPopulation, TriplePopulation
from .populations import Observed_TriplePopulation, Observed_BinaryPopulation

from .extinction import get_AV_infinity
from .trilegal import get_trilegal

